/*
 * Copyright (c) 2018 Alibaba Group Holding Ltd. All rights reserved.
 */

'use strict';

const constant = require('../../../utils/constant');
const Log = require('../../../utils/log');
const Try = require('../try');

class SystemFunction {
  constructor(functionService) {
    this.functionService = functionService;
    this.id = constant.DEFAULT_FUNCTION_ID;
  }

  start() {
    if (this.id === constant.DEFAULT_FUNCTION_ID) {
      throw new Error(`Derived class should provide its own id.`);
    }
    this._getTaskAndInvoke();
  }

  _getTaskAndInvoke() {
    var doInvoke = (err, task) => {
      var invokerContextTry = this._getInvokerContext(task.invokerContext);
      if (invokerContextTry.error) {
        Log.i(TAG, `Cannot parse the given InvokerContext: ${invokerContextTry.error}`);
        return;
      }
      var context = {
        invokerContext: {
          custom: invokerContextTry.value.custom ? invokerContextTry.value.custom : {},
        },
      };
      var postCallback = () => {
        setTimeout(() => {
          this._getTaskAndInvoke();
        }, 0);
      };
      this.handler(task.payload, context, (err, data) => {
        if (err) {
          this.functionService.postHandlerError(this.id, task.invocationId, 'Handled',
            JSON.stringify(this._getErrorObject(err)),
            (error) => postCallback());
        } else {
          this.functionService.postTaskResult(this.id, task.invocationId, data,
            (error) => postCallback());
        }
      });
    };
    this.functionService.getTask(this.id, doInvoke);
  }

  handler(event, context, callback) {
    // Derived classes should offer its own implementation.
  }

  _getErrorObject(error) {
    if (error instanceof Error) {
      return {
        errorMessage: error.message,
        errorType: error.message,
        // Do NOT return stack trace for system defined functions.
      };
    } else {
      return {
        errorMessage: error.toString(),
      }
    }
  }

  _getInvokerContext(invokerContext) {
    if (invokerContext) {
      var invokerContextTry = Try.Base64.decode(invokerContext);
      if (invokerContextTry.error) {
        return invokerContextTry;
      }
      return Try.JSON.parse(invokerContextTry.value.toString());
    } else {
      return Try.TryResult.newValue({});
    }
  }
}

module.exports = SystemFunction;