/*
 * Copyright (c) 2018 Alibaba Group Holding Ltd. All rights reserved.
 */

'use strict';

const Log = require('../../utils/log');
const TAG = '[FunctionService]';
Log.i(TAG, 'Function Compute starting...');

process.on('uncaughtException', function (err) {
  Log.e(TAG, err.stack.replace(/[\r\n]/g, ''));
  Log.i(TAG, err.stack);
});

// FUNCTION COMPUTE START
const util = require('util');
const path = require('path');
const crypto = require('crypto');
const child_process = require('child_process');
const getUUID = require('../../utils/uuid');
const wss = require('../dbus/WSServer');
const constant = require('../../utils/constant');
const fields = require('../../utils/function-fields');
const Try = require('./try');
const CredProxy = require('../dbus/dbus-proxy/CredProxy');
const credProxy = new CredProxy();
const TriggerManager = require('./triggerManager');
const ThingFunction = require('./functions/thing');
const RouterFunction = require('./functions/router');

const ROOT = require('../../const').PROJECT_PATH;

const DEFAULT_API_VERSION = 'v1';
const HEADER_AUTH_TOKEN = 'Authorization';
const HEADER_INVOCATION_ID = 'X-Fc-Invocation-Id';
const HEADER_INVOCATION_TYPE = 'X-Fc-Invocation-Type';
const HEADER_INVOKER_CONTEXT = 'X-Fc-Invoker-Context';
const HEADER_FUNCTION_ERROR = 'X-Fc-Function-Error';
const INVOCATION_TYPE_SYNC = 'Sync';
const INVOCATION_TYPE_ASYNC = 'Async';

const SYSTEM_FUNCTIONS = [
  constant.ROUTER_FUNCTION_ID,
  constant.THING_FUNCTION_ID,
];

const FUNCTION_ID_PATTERN = /^fc[a-f0-9]{32}/;
const FUNCTION_ARN_PATTERN = /^acs:fc:([a-z]{2}-[a-z]+(-\d)?)?:(\d{16})?:service:([a-zA-Z_][\w-]{0,127}):function:([a-zA-Z_][\w-]{0,127})$/;
const SERVICE_NAME_INDEX = 4;
const FUNCTION_NAME_INDEX = 5;

function res(tag, data, ws) {
  Log.i(TAG, `Responding to request "${tag}" with "${JSON.stringify(data)}".`);
  if (data.body !== undefined && data.body !== null && !(data.body instanceof Buffer)
    && Object.prototype.toString.call(data.body) !== '[object String]') {
    // var stringTry = Try.JSON.stringify(data.body);
    // if (stringTry.error) {
    //   // Ignore the error.
    //   Log.e(TAG, 'Payload in data is not JSON compatible.');
    //   return;
    // } else {
    //   Log.i(TAG, 'Stringify data is ' + stringTry.value);
    //   data.body = stringTry.value;
    // }
    data.body = JSON.stringify(data.body);
  }
  if (!data.headers) {
    // Headers should always exist.
    data.headers = {};
  }
  if (data.body) {
    data.body = Buffer.from(data.body).toString('base64');
  }
  var msg = wss.wrapResponse(tag, null, data);
  if (ws.readyState === ws.OPEN) {
    ws.send(JSON.stringify(msg));
  }
}

function message(tag, statusCode, message, ws) {
  var data = {
    statusCode,
    body: JSON.stringify({ message }),
  };
  res(tag, data, ws);
}

/**
 * Removes a specified element from the array.
 *
 * @param {Array} array the array.
 * @param {*} element the element to be removed.
 * @return {*} return the removed element if found, or undefined.
 */
function remove(array, element) {
  if (!array || array.length === 0) return;

  var index = array.indexOf(element);
  if (index === -1) return;

  var dels = array.splice(index, 1);
  return dels[0];
}

function signString(source, secret) {
  const buff = crypto.createHmac('sha256', secret).update(source, 'utf8').digest();
  return new Buffer(buff, 'binary').toString('base64');
}

function killAllFunctionProcesses() {
  child_process.exec('pkill -f cgctl');
}

/**
 * A array with maximum capacity and it discards the oldest item if pushed when it reaches
 * the maximum capacity.
 *
 * @param capacity the maximum capacity.
 * @constructor
 */
function DiscardArray(capacity) {
  if (capacity <= 0) {
    throw new Error(`Capacity must be greater than 0.`);
  }
  var storage = [];
  storage.push = function () {
    if (capacity === storage.length) {
      storage.shift();
    }
    Array.prototype.push.apply(storage, arguments);
  };
  return storage;
}

function newPostedTaskArray() {
  return new DiscardArray(200);
}

function newGotTaskResultArray() {
  return new DiscardArray(200);
}

function newPostedTaskResultArray() {
  return new DiscardArray(200);
}

function FunctionService() {
  // The pending requests from invoker to execute a task on function.
  this._postedTasks = {};
  // The pending requests from function to get task to execute.
  this._gotTasks = {};
  // The pending requests from invoker to get result of executing a task.
  this._gotTaskResults = {};
  // The task results or handler errors of function executions.
  this._postedTaskResults = {};
  // The ids of current invocations. We need the invocation list besides the invocation
  // ids in _postedTasks because the tasks in _postedTasks pop when they are processed.
  this._invocationIds = [];
  // The processes that deal with the posted tasks. For long-lived function, only one
  // process corresponding to it to deal with all tasks.
  this._taskProcesses = {};
  // The checkers that check the timeout of tasks running. System will kill the process
  // if a task running timeout.
  this._timeoutCheckers = [];
}

FunctionService.prototype.handleRequest = function (req, ws) {
  Log.i(TAG, `Handling request ${JSON.stringify(req)}.`);
  if (!req.options) {
    message(req.tag, 400, 'Options is not found.', ws);
    return;
  }
  var method = req.options.method;
  if (!method) {
    message(req.tag, 400, 'Options.method is not found.', ws);
    return;
  }
  var path = req.options.path;
  if (!path) {
    message(req.tag, 400, 'Options.path is not found.', ws);
    return;
  }
  var segments = path.split('/');
  if (method === 'GET' && segments[2] === 'credentialprovider') {
    this.handleGetCredentialRequest(req, ws);
    return;
  }
  var functionId = segments[3];
  if (functionId === undefined || functionId === null) {
    message(req.tag, 400, `Function id is not found in path.`, ws);
    return;
  }
  // For normal function, do nothing if it's formed in function id, or extract service
  // name and function name to be converted to function id if it's formed function arn
  // string.
  if (!SYSTEM_FUNCTIONS.includes(functionId) && !FUNCTION_ID_PATTERN.test(functionId)) {
    var results = functionId.match(FUNCTION_ARN_PATTERN);
    if (!results) {
      message(req.tag, 400, `Can't parse given string as a function arn: ${functionId}.`, ws);
      return;
    }
    var serviceName = results[SERVICE_NAME_INDEX];
    var functionName = results[FUNCTION_NAME_INDEX];
    if (serviceName === undefined || serviceName === null) {
      message(req.tag, 400, `Can't find required field service name in ${functionId}.`, ws);
      return;
    }
    if (functionName === undefined || functionName === null) {
      message(req.tag, 400, `Can't find required field function name in ${functionId}.`, ws);
      return;
    }
    functionId = triggerManager.getFunctionIdByName(serviceName, functionName);
    if (!functionId) {
      message(req.tag, 400, `No function found for service name ${serviceName}, function name ${functionName}.`, ws);
      return;
    }
  }
  if (method === 'GET') {
    if (segments[4] === 'task') {
      this.handleGetTaskRequest(functionId, req, ws);
    } else {
      this.handleGetTaskResultRequest(functionId, req, ws);
    }
    return;
  }
  if (method === 'POST') {
    if (segments[4] === 'task') {
      this.handlePostTaskResultOrHandlerErrorRequest(functionId, req, ws);
    } else {
      this.handlePostTaskRequest(functionId, req, ws);
    }
    return;
  }
  message(req.tag, 400, `Unknown request: "${req}".`, ws);
};

FunctionService.prototype.getTask = function (functionId, callback) {
  var gotTask = {
    callback,
    functionId,
  };
  this._getTask(functionId, gotTask);
};

FunctionService.prototype._getTask = function (functionId, gotTask) {
  var postedTask = (this._postedTasks[functionId] || []).shift();
  if (postedTask) {
    this._dispatchTask(gotTask, postedTask);
  } else {
    if (!this._gotTasks[functionId]) {
      this._gotTasks[functionId] = [];
    }
    this._gotTasks[functionId].push(gotTask);
    Log.i(TAG, `Function ${functionId} is waiting for task...`);
  }
};

FunctionService.prototype._dispatchTask = function (gotTask, postedTask) {
  var functionId = gotTask.functionId;
  Log.i(TAG, `Dispatching task "${JSON.stringify(postedTask)}" to function "${functionId}".`);

  var dispatchTask = function () {
    if (gotTask.callback) {
      // System functions, execute the task in next event.
      setTimeout(function () {
        gotTask.callback(null, postedTask);
      }, 0);
      return;
    }
    // Normal functions, return the task.
    var data = {
      statusCode: 200,
      headers: {
        [HEADER_INVOCATION_ID]: postedTask.invocationId,
        [HEADER_INVOKER_CONTEXT]: postedTask.invokerContext,
      },
      body: postedTask.payload,
    };
    res(gotTask.tag, data, gotTask.connection);
  };
  // Do not apply timeout check to system functions.
  if (SYSTEM_FUNCTIONS.includes(functionId)) {
    dispatchTask();
    return;
  }

  var self = this;
  this._getFunctionInfo(functionId, function (err, info) {
    if (err) {
      // This case will never happen.
      Log.e(TAG, `Get function info failed due to ${err}.`);
      return;
    }
    var timeout = info.timeout === undefined || info.timeout === null ? 5 : info.timeout;
    var timeoutId = setTimeout(function () {
      Log.i(TAG, `Task "${JSON.stringify(postedTask)} run timeout, kill it...".`);
      remove(self._timeoutCheckers, timeoutChecker);
      self.stopFunction(functionId, function () {/* Ignore the result */});
    }, timeout * 1000);
    var timeoutChecker = {
      functionId,
      invocationId: postedTask.invocationId,
      timeoutId,
    };
    self._timeoutCheckers.push(timeoutChecker);
    dispatchTask();
  });
};

FunctionService.prototype.handleGetCredentialRequest = function (req, ws) {
  Log.i(TAG, `Handling getting credential request for "${req.tag}."`);
  credProxy.getCredential('', function (err, credential) {
    var data;
    if (err) {
      var errString = `Got credential for request "${req.tag}" failed due to ${err}.`;
      Log.i(TAG, errString);
      data = {
        statusCode: 400,
        body: errString,
      };
    } else {
      var cred = JSON.stringify(credential);
      Log.i(TAG, `Got credential ${cred} for request "${req.tag}".`);
      data = {
        statusCode: 200,
        headers: {
          'X-Provider-Credential': cred,
        },
      };
    }
    res(req.tag, data, ws);
  });
};

FunctionService.prototype.handleGetTaskRequest = function (functionId, req, ws) {
  Log.i(TAG, `Handling getting task request for function "${functionId}".`);
  var gotTask = {
    tag: req.tag,
    connection: ws,
    functionId,
  };
  this._getTask(functionId, gotTask);
};

FunctionService.prototype.handleGetTaskResultRequest = function (functionId, req, ws) {
  Log.i(TAG, `Handling getting task result request for function "${functionId}".`);
  var headers = req.options.headers;
  if (!headers) {
    message(req.tag, 400, 'Options.header is not found.', ws);
    return;
  }
  var invocationId = headers[HEADER_INVOCATION_ID];
  if (!invocationId) {
    message(req.tag, 400, `Options.header[X-Fc-Invocation-Id] "${invocationId}" is illegal.`, ws);
    return;
  }
  if (!this._invocationIds.includes(invocationId)) {
    message(req.tag, 400, `Invocation id "${invocationId}" does not exist.`, ws);
    return;
  }
  var gotTaskResult = {
    functionId,
    invocationId,
    tag: req.tag,
    connection: ws,
  };
  this._getTaskResult(functionId, gotTaskResult);
};

FunctionService.prototype.getTaskResult = function (functionId, invocationId, callback) {
  var gotTaskResult = {
    functionId,
    invocationId,
    callback,
  };
  this._getTaskResult(functionId, gotTaskResult);
};

FunctionService.prototype._getTaskResult = function (functionId, gotTaskResult) {
  var invocationId = gotTaskResult.invocationId;
  var postedTaskResult = (this._postedTaskResults[functionId] || [])
    .find(function (postedTaskResult) {
      return postedTaskResult.invocationId === invocationId;
    });
  if (postedTaskResult) {
    remove(this._postedTaskResults[functionId], postedTaskResult);
    this._dispatchTaskResultOrHandlerError(gotTaskResult, postedTaskResult);
  } else {
    if (!this._invocationIds.includes(gotTaskResult.invocationId)) {
      // The invocation id is illegal!
      if (gotTaskResult.callback) {
        var err = new Error(`Invocation id "${invocationId}" does not exist.`);
        gotTaskResult.callback(err);
      } else {
        message(gotTaskResult.tag, 400, `Invocation id "${invocationId}" does not exist.`,
          gotTaskResult.connection);
      }
      return;
    }
    if (!this._gotTaskResults[functionId]) {
      this._gotTaskResults[functionId] = newGotTaskResultArray();
    }
    this._gotTaskResults[functionId].push(gotTaskResult);
    Log.i(TAG, `Waiting for task result from function ${functionId}...`);
  }
};

FunctionService.prototype.handlePostTaskRequest = function (functionId, req, ws) {
  Log.i(TAG, `Handling posting task request for function ${functionId}.`);
  // Check inputs...
  var headers = req.options.headers;
  if (!headers) {
    message(req.tag, 400, 'Options.header is not found.', ws);
    return;
  }
  var invocationType = headers[HEADER_INVOCATION_TYPE];
  if (invocationType !== INVOCATION_TYPE_SYNC && invocationType !== INVOCATION_TYPE_ASYNC) {
    message(req.tag, 400, 'Options.header[X-Fc-Invocation-Type] is not "' +
      INVOCATION_TYPE_SYNC + '" or "' + INVOCATION_TYPE_ASYNC + '".', ws);
    return;
  }
  var invokerContext = headers[HEADER_INVOKER_CONTEXT];
  var payload;
  var body = req.options.body;
  if (body !== undefined && body !== null) {
    payload = Buffer.from(body, 'base64');
  }
  var params = {
    functionId,
    invocationType,
    invokerContext,
    payload,
  };
  this._invokeFunction(params, function (err, invocationId) {
    if (err) {
      message(req.tag, 400, err.message, ws);
      return;
    }
    var data = {
      statusCode: invocationType === INVOCATION_TYPE_SYNC ? 200: 202,
      headers: {
        [HEADER_INVOCATION_ID]: invocationId
      },
    };
    res(req.tag, data, ws);
  });
};

FunctionService.prototype.invokeFunction = function (params, callback) {
  Log.i(TAG, `Invoke function ${params.functionId}.`);
  if (!params.functionId) {
    callback && callback(new Error(`Illegal function id ${params.functionId}.`));
    return;
  }
  if (params.invocationType !== INVOCATION_TYPE_SYNC &&
    params.invocationType !== INVOCATION_TYPE_ASYNC) {
    callback && callback(new Error(`Illegal invocation type ${params.invocationType}.`));
    return;
  }
  this._invokeFunction(params, function (err, invocationId) {
    callback && callback(err, invocationId);
  });
};

FunctionService.prototype._invokeFunction = function (params, callback) {
  Log.i(TAG, `Invoking function ${params.functionId}.`);
  // TODO Offer a better generator for invocation id.
  var invocationId = getUUID();
  // First response to the request.
  if (callback) {
    callback(null, invocationId);
  }

  // Second handle the task.
  this._invocationIds.push(invocationId);

  var postedTask = {
    invocationId,
    invokerContext: params.invokerContext,
    payload: params.payload,
  };
  var functionId = params.functionId;
  // Wake up a process if there is one waiting for a task.
  var gotTask = (this._gotTasks[functionId] || []).shift();
  if (gotTask) {
    Log.i(TAG, `Wake up function "${gotTask.functionId}" to handle the task.`);
    this._dispatchTask(gotTask, postedTask);
    return;
  }

  if (!this._postedTasks[functionId]) {
    this._postedTasks[functionId] = newPostedTaskArray();
  }
  this._postedTasks[functionId].push(postedTask);
  Log.i(TAG, `Task for function "${functionId}" has been posted: ${JSON.stringify(postedTask)}.`);

  if (SYSTEM_FUNCTIONS.includes(functionId)) {
    // System functions never need to create a process.
    return;
  }

  // Check whether it's necessary to create a new process to deal with the task.
  var isLongLived = this.isLongLivedFunction(functionId);
  var processes = this._taskProcesses[functionId];
  if ((isLongLived && (!processes || processes.length < 1)) ||
      (!isLongLived && (!processes || processes.length < 1))) {
    // Create a new process to deal with the task if it's on-demand and process
    // count < 1, or it's long-lived and no process has been created for it.
    var self = this;
    var handleError = function (err) {
      if (err) {
        self.postHandlerError(functionId, invocationId, 'Unhandled', err, null);
      }
    };
    this._getFunctionInfo(functionId, function (err, info) {
      if (err) {
        handleError(err);
      } else {
        self._startFunctionProcess(info, handleError);
      }
    });
  }
};

FunctionService.prototype.startFunction = function (functionId, callback) {
  Log.i(TAG, `Start function ${functionId}.`);
  var self = this;
  this._getFunctionInfo(functionId, function (err, info) {
    if (err) {
      callback && callback(err);
      return;
    }
    if (info.runMode !== constant.CONFIG_RUNMODE_LONGLIVE) {
      callback && callback(new Error(`Can not start non long-lived function.`));
      return;
    }
    if (self._taskProcesses[functionId] && self._taskProcesses[functionId].length === 1) {
      Log.i(TAG, `Long-lived function "${functionId}" has been started, skip.`);
      callback && callback(null);
      return;
    }
    var restart = function () {
      setTimeout(function () {
        self.startFunction(functionId, function(err, child) {
          if (err) {
            Log.i(TAG, util.format('Long-lived function "%s" restarted failed due to' +
              ' %s, try again...', functionId, err));
            restart();
          } else {
            Log.i(TAG, `Long-lived function "${functionId}" restarted successfully.`);
          }
        })
      }, 0);
    };
    var onexit = function () {
      Log.i(TAG, `Long-lived function "${functionId}" exited.`);
      if (self._taskProcesses[functionId]) {
        Log.i(TAG, `Long-lived function "${functionId}" is going to restart.`);
        restart();
      }
    };
    self._startFunctionProcess(info, function (err, child) {
      if (!err) {
        child.once('exit', onexit);
      }
      callback(err, child);
    });
  });
};

FunctionService.prototype.stopFunction = function (functionId, callback) {
  Log.i(TAG, `Stop function ${functionId}.`);
  var processes = this._taskProcesses[functionId];
  if (processes) {
    delete this._taskProcesses[functionId];
    processes.forEach(function (process) {
      process.kill('SIGTERM');
    });
  }
  callback(null);
};

FunctionService.prototype._getFunctionInfo = function (functionId, callback) {
  triggerManager.getFunctionInfo(functionId, callback);
};

FunctionService.prototype._startFunctionProcess = function (info, callback) {
  Log.i(TAG, 'Starting function process...');
  var options = {
    cwd: info.path,
  };
  var authToken = signString(info.functionId + Date.now(), getUUID());
  var variables = info[constant.CONFIG_ENV] && info[constant.CONFIG_ENV][constant.VARIABLES];
  options.env = Object.assign({}, process.env, variables);
  options.env.FUNCTION_ID = fields.buildArnString(info.regionId, null,
    info.serviceName, info.functionName);
  options.env.REGION = info.regionId;
  options.env.SERVICE_NAME = info.serviceName;
  options.env.FUNCTION_NAME = info.functionName;
  options.env.FC_FUNC_CODE_PATH = info.path;
  options.env.MY_FUNCTION_ID = options.env.FUNCTION_ID;
  options.env.ROUTER_FUNCTION_ID = constant.ROUTER_FUNCTION_ID;
  options.env.THING_FUNCTION_ID = constant.THING_FUNCTION_ID;
  options.env.AUTH_TOKEN = authToken;
  options.env.DEFFAULT_CREDENTIALS_RELATIVE_URI = '/v1/credentialprovider/';

  var command;
  var args;
  var addition;
  if (info.runtime === 'nodejs8') {
    command = 'node';
    args = [`/runtime/${info.runtime}/runtime.js`, `--handler=${info.handler}`];
    addition = `${info.path}:/runtime/nodejs8:/runtime/nodejs8/node_modules`;
    options.env.NODE_PATH = options.env.NODE_PATH ? `${addition}:${options.env.NODE_PATH}`
      : `${addition}`;
  } else if (info.runtime === 'python3') {
    command = 'python3';
    args = [`/runtime/${info.runtime}/runtime.py`, `--handler=${info.handler}`];
    addition = `${info.path}`;
    options.env.PYTHONPATH = options.env.PYTHONPATH ?
      `${addition}:${options.env.PYTHONPATH}` : `${addition}`;
  } else {
    // Unsupported runtime. Error!
    callback(new Error(`Unsupported runtime: ${info.runtime}.`));
    return;
  }
  addition = '/build-env/encrypt:/gw_lib';
  options.env.LD_LIBRARY_PATH = options.env.LD_LIBRARY_PATH ?
    `${addition}:${options.env.LD_LIBRARY_PATH}` : `${addition}`;

  var cgroupConfig = {
    config: {
      functionId: info.functionId,
      memorySize: info.memorySize,
    },
  };
  var cgroupConfigString = JSON.stringify(cgroupConfig);
  var fs = require('fs');
  var files = fs.readdirSync('/');
  var startParameters = {
    cwd: process.cwd(),
    rootfs: files,
    cgroup: cgroupConfig,
    command,
    args,
    options,
  };
  Log.i(TAG, `Start parameters: ${JSON.stringify(startParameters)}`);
  var child = child_process.spawn('/cgctl', [cgroupConfigString, command, ...args], options);
  var self = this;
  child.once('exit', function (code, signal) {
    // Remove the child.
    remove(self._taskProcesses[info.functionId], child);
  });
  child.stdout.on('data', function (data) {
    Log.i('[Function]', data);
  });
  child.stderr.on('data', function (data) {
    Log.i('[Function]', data);
  });
  if (!this._taskProcesses[info.functionId]) {
    this._taskProcesses[info.functionId] = [];
  }
  this._taskProcesses[info.functionId].push(child);
  callback(undefined, child);
};

FunctionService.prototype.handlePostTaskResultOrHandlerErrorRequest
  = function (functionId, req, ws) {
  var headers = req.options.headers;
  if (!headers) {
    message(req.tag, 400, 'Options.header is not found.', ws);
    return;
  }
  var invocationId = headers[HEADER_INVOCATION_ID];
  if (!invocationId) {
    message(req.tag, 400, `Options.header[X-Fc-Invocation-Id] "${invocationId}" is illegal.`, ws);
    return;
  }
  var error = headers[HEADER_FUNCTION_ERROR];
  if (error !== undefined && error !== null && error !== 'Handled') {
    message(req.tag, 400, `Options.header[X-Fc-Function-Error] "${error}" is illegal.`, ws);
    return;
  }
  if (!this._invocationIds.includes(invocationId)) {
    var operation = error ? 'handler error' : 'task result';
    message(req.tag, 400, `Illegal posting ${operation} request "${invocationId}".`, ws);
    return;
  }
  if (error) {
    this.handlePostHandlerErrorRequest(functionId, req, ws);
  } else {
    this.handlePostTaskResultRequest(functionId, req, ws);
  }
};

FunctionService.prototype.handlePostTaskResultRequest = function (functionId, req, ws) {
  Log.i(TAG, `Handling posting task result request for function "${functionId}".`);
  // First response to the request.
  message(req.tag, 200, 'Success', ws);

  // Second store the result, and send it to those who request to get it.
  var body = req.options.body;
  var postedTaskResult = {
    invocationId: req.options.headers[HEADER_INVOCATION_ID],
    value: body ? Buffer.from(body, 'base64') : body,
  };
  this._handleTaskResultOrHandlerError(functionId, postedTaskResult);
};

FunctionService.prototype.postTaskResult = function (functionId, invocationId, value,
                                                     callback) {
  Log.i(TAG, `Handling posting task result request for function "${functionId}".`);
  // First response to the request.
  if (callback) {
    callback(null);
  }

  // Second dispatch the result if some one is waiting for it, or store it.
  var postedTaskResult = {
    invocationId,
    value,
  };
  this._handleTaskResultOrHandlerError(functionId, postedTaskResult);
};

FunctionService.prototype._handleTaskResultOrHandlerError = function (functionId, result) {
  Log.i(TAG, `Handling task result or handler error for function "${functionId}".`);
  // Cleanup the timeout checker if it is not system function.
  var index;
  if (!SYSTEM_FUNCTIONS.includes(functionId)) {
    for (index = this._timeoutCheckers.length - 1; index >= 0; index--) {
      var timeoutChecker = this._timeoutCheckers[index];
      if (timeoutChecker.functionId === functionId
        && timeoutChecker.invocationId === result.invocationId) {
        Log.i(TAG, `Clear timeout checker "${timeoutChecker.timeoutId}".`);
        clearTimeout(timeoutChecker.timeoutId);
        this._timeoutCheckers.splice(index, 1);
      }
    }
  }
  var found = false;
  var gotTaskResults = this._gotTaskResults[functionId];
  if (gotTaskResults) {
    for (index = gotTaskResults.length - 1; index >= 0; index--) {
      var gotTaskResult = gotTaskResults[index];
      if (gotTaskResult.invocationId === result.invocationId) {
        found = true;
        gotTaskResults.splice(index, 1);
        this._dispatchTaskResultOrHandlerError(gotTaskResult, result);
      }
    }
  }
  if (!found) {
    if (!this._postedTaskResults[functionId]) {
      this._postedTaskResults[functionId] = new newPostedTaskResultArray();
    }
    this._postedTaskResults[functionId].push(result);
    var message = result.error ? 'Handler error' : 'Task result';
    Log.i(TAG, `${message} for function ${functionId} has been posted.`);
  }
};

FunctionService.prototype._dispatchTaskResultOrHandlerError = function (gotTaskResult, postedTaskResult) {
  var message = postedTaskResult.error ? 'handler error' : 'task result';
  Log.i(TAG, `Dispatching ${message} for function ${gotTaskResult.functionId}...`);
  var data;
  if (gotTaskResult.callback) {
    data = {
      value: postedTaskResult.value,
    };
    if (postedTaskResult.error) {
      data.error = postedTaskResult.error;
    }
    gotTaskResult.callback(null, data);
  } else {
    data = {
      statusCode: 200,
      body: postedTaskResult.value,
    };
    if (postedTaskResult.error) {
      data.headers = {
        [HEADER_FUNCTION_ERROR]: postedTaskResult.error,
      };
    }
    res(gotTaskResult.tag, data, gotTaskResult.connection);
  }
  remove(this._invocationIds, postedTaskResult.invocationId);
};

FunctionService.prototype.handlePostHandlerErrorRequest = function (functionId, req, ws) {
  Log.i(TAG, `Handling posting handler error request for function "${functionId}".`);
  if (!req.options.body) {
    message(req.tag, 400, 'Error body is not found.', ws);
    return;
  }

  if (req.options.headers[HEADER_FUNCTION_ERROR] !== 'Handled') {
    message(req.tag, 400, 'Headers[X-Fc-Function-Error] should be "Handled".', ws);
    return;
  }

  // First response to the request.
  message(req.tag, 200, 'Success', ws);
  // Second store the result, and send it to those who request to get it.
  var body = req.options.body;
  var postedHandlerError = {
    invocationId: req.options.headers[HEADER_INVOCATION_ID],
    error: req.options.headers[HEADER_FUNCTION_ERROR],
    value: body ? Buffer.from(body, 'base64').toString() : body,
  };
  this._handleTaskResultOrHandlerError(functionId, postedHandlerError);
};

FunctionService.prototype.postHandlerError = function (functionId, invocationId, type,
                                                       error, callback) {
  Log.i(TAG, `Handling posting handler error request for function "${functionId}".`);
  // First response to the request.
  if (callback) {
    callback(null);
  }

  // Second dispatch the result if some one is waiting for it, or store it.
  var postedHandlerError = {
    invocationId,
    error: type,
    value: error,
  };
  this._handleTaskResultOrHandlerError(functionId, postedHandlerError);
};

FunctionService.prototype.isLongLivedFunction = function (functionId) {
  return triggerManager.isLongLivedFunction(functionId);
};

FunctionService.prototype.dumpStates = function () {
  var dumpStates =
    `
    ------------------- Dump States --------------------
    Invocations: ${JSON.stringify(this._invocationIds)},
    Processes: ${JSON.stringify(this._taskProcesses)},
    PostedTasks: ${JSON.stringify(this._postedTasks)},
    PostedTaskResults: ${JSON.stringify(this._postedTaskResults)},
    ----------------------------------------------------
    `;
  Log.i(TAG, dumpStates);
};

killAllFunctionProcesses();
const functionService = new FunctionService();
wss.setRequestHandler(functionService);
new ThingFunction(functionService).start();
new RouterFunction(functionService).start();
const triggerManager = new TriggerManager(functionService);

// FUNCTION COMPUTE END

module.exports = functionService;
