const Tag = require('../../utils/constant');
const log = require('../../utils/log');
const Trigger = require('./trigger.js');
const util = require('util');
var condition = require("./condition");
const TAG = "IFTTT_scene";
// var Operation = require('./operation.js');

var Scene = function (scheduler, eventBus, statemachine) {
    this.sceneid = 0; // 场景编号
    this.enabled = true; // 是否被允许
    this.operation = null; // 逻辑符，and/or
    this.triggers = []; // 条件列表
    this.jobInfo = {}; // 与jobmanager沟通的信息
    this.scheduler = scheduler;
    this.eventBus = eventBus;
    this.statemachine = statemachine;
}

Scene.prototype.parseScene = function (sceneObj) {
    log.i(TAG, "sceneObj = " + JSON.stringify(sceneObj));
    var legal = sceneObj.hasOwnProperty(Tag.ID) &&
        // sceneObj.hasOwnProperty(Tag.TRIGGER) &&
        sceneObj.hasOwnProperty(Tag.ACTION);
    if (legal) {
        this.sceneObj = sceneObj;
        this.sceneid = sceneObj[Tag.ID];
        if (sceneObj.hasOwnProperty(Tag.CONDITION)) {
            var conditionObject = sceneObj[Tag.CONDITION];
            if (conditionObject['items'] != undefined) {
                this.sceneObj[Tag.CONDITION] = conditionObject['items'];
            } else {
                var items = [];
                items.push(conditionObject);
                this.sceneObj[Tag.CONDITION] = items;
            }
        }
        if (sceneObj.hasOwnProperty(Tag.ENABLED)) {
            this.enabled = sceneObj[Tag.ENABLED];
        }
        legal = this.parseTriggers(sceneObj[Tag.TRIGGER]) && legal;
    }
    return legal;
};

Scene.prototype.parseTriggers = function(triggers) {
    if(!triggers || Object.keys(triggers).length === 0) {
        this.operation = this.operation = Tag.LogicOr;
        return true;
    }
    var self = this;
    this.operation = triggers[Tag.URI];
    var condilist = triggers[Tag.ITEMS] || [];
    // items是空的时候，意味着只有一个条件
    if (condilist.length === 0) {
        log.i(util.format("[IFTTT::Scene::parseTriggers] no trigger items specified, there is only one trigger."));
        var condi = {};
        condi[Tag.URI] = triggers[Tag.URI];
        condi[Tag.PARAMS] = triggers[Tag.PARAMS];
        this.operation = Tag.LogicOr; // 只有一个条件的时候，置为or逻辑，与多个条件时兼容
        condilist.push(condi);
    }
    var result = false;
    condilist.forEach(function(condi) {
        var con = new Trigger(condi, self.sceneid, self.eventBus, self.statemachine);
        result = con.start();
        if (result) {
            self.triggers.push(con);
        }
    });
    return result;
};

Scene.prototype.propertiesChanged = function(info) {
    if (!this.enabled) {
        return;
    }
    var canExecute = false;
    if (this.operation === Tag.LogicOr) {
        canExecute = this.triggers.some(function(condi) {
            return condi.onTrigger(info);
        });
    } else if (this.operation === Tag.LogicAnd) {
        canExecute = this.triggers.every(function(condi) {
            // And逻辑下，需要检查是否本次被触发/或者曾经触发过
            return condi.onTrigger(info) || condi.isFullFilled();
        });
    } else {
        log.i(util.format("[ERRPR] Scene '%s' trigger operation '%s' is not supported.",
            this.sceneid, this.operation));
        canExecute = false;
    }

    log.i('Scene:', this.sceneid, 'propertiesChanged triggered, triggered canExecute:', canExecute);
    var self = this;
    if (canExecute) {
        log.i(TAG, " condition = " + JSON.stringify((this.sceneObj)[Tag.CONDITION]));
        // log.i('Scene: propertiesChanged triggered, isConditionMeet begin condition size = ' + (this.sceneObj)[Tag.CONDITION].length);
        condition.isConditionsMeeted((this.sceneObj)[Tag.CONDITION], function (err, satisfy) {
            log.i(TAG, this.sceneid, 'propertiesChanged isConditionsMeeted err =', err, 'satisfy =', satisfy);
            if(!err && satisfy) {
                self.executeJob();
            }
        });
    }
};

Scene.prototype.timerTriggered = function (info) {
    // log.i(TAG, "timerTriggered begin info = " + JSON.stringify(info));
    if (this.sceneid !== info.sceneid || !this.enabled) return undefined;
    var self = this;
    condition.isConditionsMeeted((this.sceneObj)[Tag.CONDITION], function (err, satisfy) {
        log.i(TAG, this.sceneid, 'timerTriggered isConditionsMeeted err =', err, 'satisfy =', satisfy);
        if(!err && satisfy) {
            self.executeJob();
        }
    });
};

Scene.prototype.eventHappened = function(info) {
    log.i("Scene eventHappened_" + this.sceneid + this.triggers.length + ", this.enabled = " + this.enabled);
    if (!this.enabled) {
        return;
    }

    var canExecute = false;
    if (this.operation === Tag.LogicOr) {
        canExecute = this.triggers.some(function(condi) {
            return condi.eventHappened(info);
        });
    } else if (this.operation === Tag.LogicAnd) {
        canExecute = this.triggers.every(function(condi) {
            // And逻辑下，需要检查是否本次被触发/或者曾经触发过
            return condi.eventHappened(info) || condi.isFullFilled();
        });
    } else {
        log.i(util.format("[ERRPR] Scene '%s' trigger operation '%s' is not supported.",
            this.sceneid, this.operation));
        canExecute = false;
    }

    log.i('Scene: eventHappened triggered, canExecute:', canExecute);
    if (canExecute) {
        var self = this;
        condition.isConditionsMeeted((this.sceneObj)[Tag.CONDITION], function (err, satisfy) {
            log.i(TAG, 'eventHappened isConditionsMeeted err =', err, 'satisfy =', satisfy);
            if(!err && satisfy) {
                self.executeJob();
            }
        });
       // this.executeJob();
    }
};

Scene.prototype.executeJob = function () {
    log.i('Scene::executeJob sid =', this.sceneid);
    if (this.scheduler) {
        this.scheduler.schedule("ifttt", this.jobInfo);
    }
};

Scene.prototype.stop = function () {
    var stopped = true;
    this.triggers.forEach(function(condi) {
        stopped = condi.stop() && stopped;
    });
    return stopped;
};

module.exports = Scene;
