var Group = require("./Group");
var Log = require("./log");
const fs = require("fs");
const path = require("path");
const constant = require("../../utils/constant");
const TAG = "developer_linkedge";

/**
 * 根据event 中的groupId 生成分组对象
 * @param {Buffer} event
 * @return {Group} group
 */
exports.getGroupByEvent = function (eventBuffer) {
    if (arguments.length != 1 || !eventBuffer) {
        return new Error('params error');
    }
    var eventStr = eventBuffer.toString();
    Log.LOGI(TAG, "getGroupByEvent event = " + eventStr);
    var event = JSON.parse(eventStr);
    if (event && event.provider && event.provider.groupId) {
        var groupId = event.provider.groupId;
        return new Group(groupId);
    }
    return null;
};

/**
 * 根据groupId 生成分组设备对象
 * @param groupId
 * @return {Group}
 */
exports.getGroupById = function(groupId) {
    Log.LOGI(TAG, "getGroupById groupId = " + groupId);
    if(arguments.length != 1 || !groupId || typeof groupId != 'number' || groupId < 0) {
        return new Error('params error');
    }
    return new Group(groupId);
};

/**
 * 获取当前用户目录下面分组对象
 * @return {Group}
 */
exports.getGroup = function() {
    Log.LOGI(TAG, "getGroup begin...");
    var groupAPI =  require("./api-group");
    var configPath = groupAPI.getConfigPath();
    var fileNames = fs.readdirSync(configPath);
    var groupId;
    if(fileNames.length === 1) {
        var filePath = path.join(configPath, fileNames[0], constant.jobIndex);
        var content = fs.readFileSync(filePath, "utf-8");
        try {
            var confObject = JSON.parse(content);
            groupId = confObject.groupId;
        } catch (e) {
            groupId = -1;
            Log.LOGE(TAG, 'groupConfig :', filePath, 'json Exception =',e.message);
        }
    } else {
        groupId = -1;
        Log.LOGE(TAG, "something wrong with groupConfig , configName list = " + fileNames);
    }
    if(groupId == -1) {
        return new Error('something wrong with groupConfig');
    } else {
        return new Group(groupId);
    }
};

/**
 * 根据event中的groupId 和 tagKey 找出对应的tag列表
 * "event":{
"type":"deviceEvent",
"payload":{
"id":"obj1_abcde",
"eventName":"post",
"params":{
"num":250
},
"eventType":1
},
"provider":{
"productKey":"obj1",
"deviceName":"abcde",
"groupId":11
}
}
 * @param {Buffer} event
 * @param {String} Tagkey
 * @param {Function} callback(Error err, List<Tag> tags)
 */
exports.getTag = function(event, key, callback) {
    Log.LOGI(TAG, "getTag begin key =", key);
    if(arguments.length != 3 || typeof arguments[2] != 'function') {
        Log.LOGE(TAG, "arguments[2] must be callback");
        throw new Error("arguments[2] must be callback");
        return;
    }
    if(!event || !key) {
        return callback(new Error('params error'));
    }
    Log.LOGI(TAG, "getTag event =", event.toString());
    var group = module.exports.getGroupByEvent(event);
    if(group) {
        group.getTagList(key, function (err, list) {
            return callback(err, list);
        });
    } else {
        return callback(new Error('group object is null, event =' + JSON.stringify(event.toString())));
    }
};

/**
 * 根据key,value 生成Tag 对象
 * @param key
 * @param value
 * @return {object} tag
 * @constructor
 */
exports.createTag = function(key, value) {
    if(arguments.length != 2) {
        return new Error('params error');
    }
    var result = {};
    result[key] = value;
    return result;
};

exports.publish = function (topic, payload, callback) {
    Log.LOGI(TAG, "publish functionId =" + process.env.FUNCTION_ID);
    if(arguments.length != 3 || typeof arguments[2] != 'function') {
        Log.LOGE(TAG, "arguments[2] must be callback");
        throw new Error("arguments[2] must be callback");
        return;
    }
    if(typeof topic != 'string' || typeof payload != 'string') {
	return callback(new Error('params type err, should string.'));
    }
    var runMode = process.env.RUN_MODE;
    var _interface;
    if(runMode === constant.CONFIG_RUNMODE_LONGLIVE) {
        _interface = "iot.fc.FCLong_"; //iot.fc.FCLong_{id}
    } else {
        _interface = "iot.fc.FCShort_"; //iot.fc.FCShort_{id}
    }
    _interface += process.env.FUNCTION_ID;
    if(typeof payload !== 'string') {
        payload = JSON.stringify(payload);
    }
    const ws = require("../ws");
    ws.publish(constant.FILTER_BUS_NAME, _interface, topic, "fc_event", payload, callback);
};

exports.log = require("./log");
