'use strict';
const fs = require("fs");
const path = require("path");
const Log = require('../developer/log');
const constant = require("../../utils/constant");
const TAG = "[loader_Script] ";
const { PROJECT_PATH } = require('../../const');
/**
 * job
 {
   "jobId": 0,
   "ackCode" : 10201,
   "path": "",
   "type" : "xxx",
   "executor" : "yyy",
   "jobstatus": ""  //比原来job多的一个字段
   }
 * @param callback
 * @constructor
 */
var Script = function (job, jobConfig) {
    Log.LOGI(TAG, "Script construct begin jobid = ", job.jobId);
    this.jobPath = job.path;
    this.params = getParams(job);
    this.jobConfig = jobConfig;
    try {
        if(jobConfig && jobConfig.hasOwnProperty(constant.CONFIG_CONFIG) && (jobConfig[constant.CONFIG_CONFIG]).hasOwnProperty(constant.CONFIG_TIMEOUT)) {
            this.timeLimit = jobConfig[constant.CONFIG_CONFIG][constant.CONFIG_TIMEOUT] * 1000;
        } else {
            this.timeLimit = constant.DEFAULT_TIME_OUT;
        }
        this.runMode = jobConfig[constant.CONFIG_CONFIG][constant.CONFIG_RUNMODE];
        Log.LOGI(TAG, "Script timeout " + this.timeLimit, "runMode ", this.runMode);
    } catch (err) {
        Log.LOGE(TAG, "Script construct err = ", err);
    }
};

Script.prototype.requireCode = function () {
    var fileBegin = constant.jobIndex;
    var userExports;
    try {
        var codeBegin = this.jobConfig[constant.CONFIG_CONFIG][constant.CONFIG_HANDLER];
        fileBegin = codeBegin.split(".")[0];
    } catch (err) {
        Log.LOGE(TAG, "runHandler err = " + err);
    }
    var code = path.join(this.jobPath, fileBegin + ".js");
    if (!fs.existsSync(code)) {
        return new Error('can`t find your code', code);
    }
    try {
        userExports = require(code);
    } catch (err) {
        return err;
    }
    return userExports;
};

/**
 * 开始执行脚本
 * @param jobPath 脚本路径
 */
Script.prototype.runHandler = function (callback) {
    var jobPath = this.jobPath;
    Log.LOGI(TAG, "runHandler begin jobPath =" + jobPath);
    if (jobPath && jobPath.startsWith("./")) {
        jobPath = path.join(PROJECT_PATH, jobPath.split("./")[1]);
    }
    var userExports;
    this.main = null;
    var funcBegin = "handler";
    var fileBegin = constant.jobIndex;
    try {
        var codeBegin = this.jobConfig[constant.CONFIG_CONFIG][constant.CONFIG_HANDLER];
        funcBegin = codeBegin.split(".")[1];
        fileBegin = codeBegin.split(".")[0];
    } catch (err) {
        Log.LOGE(TAG, "runHandler err = " + err);
    }
    var code = path.join(jobPath, fileBegin + ".js");
    if (!fs.existsSync(code)) {
        return callback(new Error('can`t find your code'));
    }
    userExports = this.requireCode();
    if(Object.prototype.toString.call(userExports) === '[object Error]') {
        return callback(userExports);
    }
    this.main = getFuncMain(userExports, funcBegin);
    if (!this.params) {
        return callback(new Error('[Error] your script params is null'));
    }
    if (!this.main) {
        return callback(new Error('[Error] your script don`t follow our framework rules'));
    }
    this.executeHandler(function (err, ret) {
        callback(err, ret);
    });
};

Script.prototype.refreshParams = function (event, context) {
    this.params.event = event;
    this.params.context = context;
};

/**
 * 执行handler入口
 */
Script.prototype.executeHandler = function (callback) {
    Log.LOGI(TAG, "executeHandler begin");
    //深拷贝this.config,因为执行时候config 对应的value : cloudId会被替换成真正的实例化对象
    var paramsObject= JSON.parse(JSON.stringify(this.params));
    Log.LOGI(TAG, "params = " + JSON.stringify(paramsObject));
    try {
        this.execute(this.timeLimit, this.main, paramsObject, function (err, ret) {
            Log.LOGI(TAG, "executeHandler result : err = " + err);
            callback(err, ret);
        });
    } catch (err) {
        Log.LOGE(TAG, "executeHandler catch err = " + err);
        callback(err);
    }
};

/**
 * 最小执行单元
 * @param {Function} main
 * @param {JsObject} paramsRoot 带的入参
 * @param callback
 */
Script.prototype.execute = function(timeLimit, main, paramsRoot, callback) {
    Log.LOGI(TAG, "execute begin");
    const userParams = getFnArgNames(main);
    var params = [];
    var hasReturn = false;
    Log.LOGI(TAG, "userParams=" + userParams + ", paramsRoot=" + JSON.stringify(paramsRoot));
    var func = function (err, ret) {
        hasReturn = true;
        callback(err, ret);
    };
    for(var paramKey of userParams) {
        if(paramKey === "callback") {
            params.push(func);
        }
        if(paramKey === 'event') {
            params.push(Buffer.from(JSON.stringify(paramsRoot[paramKey])));
        } else {
            params.push(paramsRoot[paramKey]);
        }
    }
    try {
        main(...params);
        // 不调用callback的不给执行完立马返回，而是将会挂超时才结束退出
        // return callback(null, "user no return");
        Log.LOGI(TAG, "execute runMode :", this.runMode);
        var processTimeout = setTimeout(function () {
            Log.LOGI(TAG, "handler finished with timeLimit ", timeLimit);
            if(!hasReturn) {
                return callback(new Error('FC handler timeout'));
            }
            clearTimeout(processTimeout);
        }, timeLimit);
    } catch (err) {
        Log.LOGE(TAG, "execute err = " + err);
        return callback(err);
    }
};

/**
 * 获取用户exports内容中的main函数
 * @param userExports
 * @return {*}
 */
function getFuncMain(userExports, entrance) {
    if (typeof userExports === 'function') {
        var main = userExports;
    }
    if (typeof userExports === 'object') {
        var main = userExports[entrance];
    }
    Log.LOGI(TAG, "user handler = " + main);
    return main;
};

function getParams(jsonObject) {
    var root = {};
    // root.event = new Buffer(JSON.stringify(jsonObject.event));
    root.event = jsonObject.event;
    root.context = jsonObject.context;
    return root;
}

/**
 * 获取一个函数的入参参数列表
 * @param fn
 * @return {Array}
 */
function getFnArgNames(fn) {
    if (typeof fn !== 'object' && typeof fn !== 'function') return;
    const COMMENTS = /((\/\/.*$)|(\/\*[\s\S]*?\*\/))/mg;
    const DEFAULT_PARAMS = /=[^,)]+/mg;
    const FAT_ARROWS = /=>.*$/mg;
    var code = fn.prototype ? fn.prototype.constructor.toString() : fn.toString();
    code = code
        .replace(COMMENTS, '')
        .replace(FAT_ARROWS, '')
        .replace(DEFAULT_PARAMS, '');
    const result = code.slice(code.indexOf('(') + 1, code.indexOf(')')).match(/([^\s,]+)/g);
    return result === null ? [] : result;
}

function isCalledCallback(fn) {
    if (typeof fn !== 'object' && typeof fn !== 'function') return;
    var code = fn.prototype ? fn.prototype.constructor.toString() : fn.toString();
    if(code.indexOf('callback(') != -1 || code.indexOf('callback (') != -1) {
        return true;
    }
    return false;
}

module.exports = Script;
